!!$=head1 NAME
!!$Elastic modeling
!!$
!!$=head1 DESCRIPTION
!!$
!!$ 2D modeling program that creates elastic data from a set of 
!!$ velocities (vp,vs) and density models or Lame (lambda, mu)
!!$ and density models. Outputs data in the form of particle
!!$ velocities, normal and shear stresses.
!!$
!!$=head1 USAGE
!!$
!!$elastic.x  <input 1 >/dev/null [PARAMETERS]
!!$
!!$=head1 INPUT PARAMETERS
!!$
!!$=over 3
!!$
!!$=item lame
!!$
!!$0 for vp,vs,rho
!!$
!!$1 for lambda,mu,rho
!!$
!!$=item input 1
!!$
!!$if lame=0, .H file of Vp model in m/s
!!$
!!$if lame=1, .H file of lambda model in MPa
!!$
!!$=item mu
!!$
!!$.H file of Mu model in MPa (for lame=1)
!!$
!!$=item vs
!!$
!!$.H file of Vs model in m/s (for lame=0)
!!$
!!$=item rho
!!$
!!$.H file of density model in kg/m3
!!$
!!$=item Vx_rec
!!$
!!$.H file of velocity x data
!!$
!!$=item Vz_rec
!!$
!!$.H file of velocity z data
!!$
!!$=item sigmaxx_rec
!!$
!!$.H file of normal stress in x data
!!$
!!$=item sigmazz_rec
!!$
!!$.H file of normal stress in z data
!!$
!!$=item sigmaxz_rec
!!$
!!$.H file of shear stress data
!!$
!!$=item xsource
!!$
!!$position of the source in the x coordinate in grid points
!!$
!!$=item zsource
!!$
!!$position of the source in the z coordinate in grid points
!!$
!!$=item Ttime
!!$
!!$total propagation time in seconds
!!$
!!$=item fp
!!$
!!$peak frequency of the Ricker type source
!!$
!!$=item abc
!!$
!!$width of absorbing boundary condition in grid points
!!$
!!$=item zrec
!!$
!!$depth of receiver line in grid points
!!$
!!$=back
!!$
!!$=head1 OPTIONAL PARAMETERS
!!$
!!$=over 3
!!$
!!$=item dtseis
!!$
!!$sampling of output seismogram in seconds [default dtseis=0.002]
!!$
!!$=item snap
!!$
!!$if 1, saves all data wavefields at every dtseis sample [default snap=0]
!!$
!!$=item rec_source
!!$
!!$if 1, reads data as input for the source use vx_in, vz_in, sigmaxx_in, sigmazz_in
!!$and sigmaxz_in to choose the .H files for input [default rec_source=0]
!!$
!!$=item dtflag
!!$
!!$if 1, reads time.txt to choose the time sampling for propagation and ratio between
!!$time sampling and seismogram sampling, instead of calculating them from stability
!!$condition [default dtflag=0]  
!!$
!!$=back
!!$
!!$=head1 ABOUT
!!$Program created by Gustavo Catao Alves at
!!$the Stanford Exploration Project
!!$
!!$Updated in March, 3rd, 2016
!!$
!!$=cut

! Main program to execute elastic modeling
program elastic_program_rec

  use sep
  use elastic_mod
  use seis_mod_rtm
  use random_mod

  use omp_lib

  implicit none

  integer                           :: nz,nx,abc,rec_length,ntseis,dtratio,data_comp
  integer*8							:: nt
  integer                           :: stat, ix, iz,rec_list_length,source_type
  real                              :: dz,oz,dx,ox,dt,fp,Tt,ot,dtseis,xsource,zsource,zrec,xrec
  real, allocatable, dimension(:,:,:) :: bkg_model 
  real, allocatable, dimension(:)   :: rec_data
  real                              :: lval,mval,rval,vtemp,vmax
  logical                           :: lame,dtflag,rec_source,snap,write_bkg,source_sig,rec_list,absorbingBound,freeSurface,adj

  !Variables to measure runtime
  real*8		:: start_time,end_time
  integer*8		:: long1,long2

  !OMP settings
  integer                               :: node, nodes
  integer, allocatable, dimension(:)	:: b,e
  start_time = omp_get_wtime()

  call sep_init(SOURCE)

  ! Read SEP header parameter
  ! Model parameters
  call from_history("n1",nz)
  call from_history("o1",oz)
  call from_history("d1",dz)
  call from_history("n2",nx)
  call from_history("o2",ox)
  call from_history("d2",dx)

  call from_param("xsource",xsource)       !position of source in x real
  call from_param("zsource",zsource)       !position of source in z real
  call from_param("Ttime",Tt)              !total propagation time
  call from_param("ot",ot,0.0)             !start of seismogram, default=0
  call from_param("fp",fp,25.0)            !Ricker source peak freq.
  call from_param("abc",abc,0)             !# of absorbing border samples
  call from_param("zrec",zrec)             !receivers depth
  call from_param("xrec",xrec)             !receivers first position in x (meters)
  call from_param("dtseis",dtseis,0.002)   !seismogram sampling, default=2ms
  call from_param("lame",lame)             !true for Lame parameters, false for velocities
  call from_param("rec_source",rec_source,.false.)  !this option injects a receiver file into the forward propagation
  call from_param("snap",snap,.false.)     !True if you want to record a movie of the wavefield
  call from_param("randomBound",randomBound,.true.)
  call from_param("absorbingBound",absorbingBound,.false.)
  call from_param("freeSurface",freeSurface,.false.) !If True, source is not shift down by abc size
  call from_param("write_bkg",write_bkg,.false.) !true if you wanna write the wavefield at the last 2 time steps
  call from_param("nodes",nodes,16)		   !number of nodes for OMP threads
  call from_param("source_type",source_type,1) !1=explosive in stress,2=vertical,3=horizontal,4=explosive in force
  call from_param("source_sig",source_sig,.false.) !If true, read the source wavelet from file
  call from_param("rec_list",rec_list,.false.) !If true, read receiver X positions from file (Z is still fixed)
  call from_param("rec_length",rec_length,0) !If 0, will put receivers at every x.
  call from_param("adj",adj,.false.) !If true, runs a wavelet inversion step
  call from_param("data_comp",data_comp,0)   !0 writes all components
											 !3 writes Vx
											 !2 writes Vz
											 !1 writes Hydrophone

  !In case dt needs to be fixed
  call from_param("dtflag",dtflag,.false.)        !TRUE if you want to read dt and dtratio from file

  if(dtflag) then
    open(unit=20,file="time.txt")
    read(20,*)dt,dtratio
    close(20)
  else
    dt=0
    dtratio=0
  endif

  !OMP
  call omp_set_num_threads(nodes)

  !Division of the model domain in threads
  allocate(b(nodes))
  allocate(e(nodes))
  
  b(1) = 1
  do ix=1,nodes-1
    e(ix) = nx/nodes
    e(ix) = ix*e(ix)
    b(ix+1) = e(ix)
  enddo
  e(nodes) = nx

  !Random Boundaries
  call randomBoundInit()
  !Background model allocation
  allocate(bkg_model(nz,nx,3))
  ! Initialize values
  bkg_model = 0.
  call sep_read(bkg_model)  !Should be ordered as density, vp and vs
  							!or density, lambda and mu
  if(randomBound) then
    if(Lame) then
      call randomLameBound(bkg_model(:,:,2),bkg_model(:,:,3),bkg_model(:,:,1))
	else
	  call randomVelBound(bkg_model(:,:,2),bkg_model(:,:,3),bkg_model(:,:,1))
	endif
  endif
  vtemp=0.
  vmax=0.
  if (lame) then
    do ix=1,nx
      do iz=1,nz
        rval=bkg_model(iz,ix,1)
        lval=1.0E6*bkg_model(iz,ix,2)
        mval=1.0E6*bkg_model(iz,ix,3)
        vtemp=sqrt((lval+2.0*mval)/rval)
        if (vtemp.gt.vmax) vmax=vtemp
	  enddo
    enddo
  else
    vmax = maxval(bkg_model(:,:,2))
  endif

 	call to_history("n1", nz, "Dat/random_model.H")
    call to_history("o1", 0., "Dat/random_model.H")
    call to_history("d1", dz, "Dat/random_model.H")
    call to_history("n2", nx, "Dat/random_model.H")
    call to_history("o2", 0., "Dat/random_model.H")
    call to_history("d2", dx, "Dat/random_model.H")
    call to_history("n3", 3, "Dat/random_model.H")
    call to_history("o3", 0., "Dat/random_model.H")
    call to_history("d3", 1., "Dat/random_model.H")
    call sep_write(bkg_model,"Dat/random_model.H")


  ! dt from the Courant stability condition => C=(vx*dt)/dx+(vz*dt)dz<1
  ! For the simple case of dz=dx and using Virieux(1986) relation:
  ntseis = ceiling(Tt/dtseis) + 1
  if (dt.eq.0.) dt = 0.4*dx/(sqrt(2.0)*vmax) !the 0.4 is arbitrary, but it was diverging otherwise
  if (dtratio.eq.0) dtratio = ceiling(dtseis/dt)
  dt = dtseis/dtratio
  !nt = Tt/dt
  nt = 2*Tt/dt !The 2 appears because in this implementation, each time loop
               !advances only dt/2

  if(rec_length.eq.0) rec_length = nx-2*abc
  if(rec_list) call from_aux ("rec_file","n1", rec_length)

  long1=5*rec_length*ntseis
  allocate(rec_data(long1))
  rec_data = 0.
  if(adj) then
    call seismo_adj(rec_length,1.0,ntseis,dtseis,rec_data)
  endif


  ! Write outputs to the terminal
  write(0,*)"============================================================="
  write(0,*)"BEGIN OF SUBROUTINE CALLS"
  write(0,*)"============================================================="
  write(0,*)"Values of the stability conditions:"
  write(0,*)"Max Vp=",vmax
  write(0,*)"dx=dy=",dx
  write(0,*)"dt=",dt
  write(0,*)"lame=",lame
  write(0,*)"dtseis=",dtseis
  write(0,*)"ntseis=",ntseis
  write(0,*)"dtratio=",dtratio
  write(0,*)"rec_length=",rec_length
  write(0,*)"============================================================="
  write(0,*)"MEMORY ALLOCATION FOR MODELS AND DATA"
  write(0,*)"============================================================="
  write(0,*)"BACKGROUND MODEL"
    write(0,*)"Parameter model",size(bkg_model)*4/(1024),"KB"
  write(0,*)"DATA"
    write(0,*)"Receiver data",size(rec_data)*4/(1024),"KB"
  write(0,*)"============================================================="

  ! Correction of source position for force sources (Vx and Vz grids are shifted from
  ! sigmaxx,zz grid by half a space sample
  !if(source_type.eq.2) zsource = zsource + (dz/2)
  !if(source_type.eq.3) xsource = xsource + (dx/2)

  ! Initialize the elastic modeling
  call elastic_init(adj,nx,dx,nz,dz,dtseis,ntseis,rec_length,rec_list,zrec,xrec,xsource,zsource,source_type,data_comp,dt,nt,fp,abc,lame,snap,write_bkg,absorbingBound,freeSurface,bkg_model(:,:,1),bkg_model(:,:,2),bkg_model(:,:,3),b,e)
  stat = elastic_op(rec_data)
  if (stat) then
    write(0,*) "ERROR ON ELASTIC_OP FUNCTION CALL ==> EXITING"
    stop
  else
    write(0,*) "PROPAGATION MODULE ==> OK"
  endif

  if (adj==.false.) then
    if(rec_list)then
   	  write(0,*) "WRITING SEISMOGRAM ONLY AT RECEIVER LOCATIONS"
  	  call seismo(rec_length,1.0,ntseis,dtseis,rec_data,data_comp)
    else
   	  write(0,*) "RECEIVER POSITIONS NOT DEFINED FROM LIST"
  	  call seismo(rec_length,1.0,ntseis,dtseis,rec_data,data_comp)
    endif
    write(0,*) "WRITE SEISMOGRAM ==> OK"
  endif

  call sep_close()

  end_time = omp_get_wtime()

  open(unit=20,file="time.txt")
  write(20,*)dt,dtratio
  close(20)

  ! Write outputs to the terminal
  write(0,*)"============================================================="
  write(0,*)"TOTAL ELAPSED TIME"
  write(0,*)"============================================================="
  write(0,*)"Total=",end_time - start_time, "seconds."
  write(0,*)"============================================================="

  deallocate(b,e,bkg_model,rec_data)

  call exit(0)

end program
