! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! May, 12th, 2015

! This module creates the density and Lame
! background models in the staggered grid.

module model_mod

  implicit none

  integer, private :: nz,nx
  real, private    :: dx
  real, pointer, dimension(:,:),private :: m1,m2,m3,m1_x,m1_z
  real, pointer, dimension(:,:),private :: dm1,dm1_x,dm1_z,dm2,dm3

  contains

  subroutine model_init(nz_in,nx_in,dx_in)
    real                        :: dx_in
    integer                     :: nx_in, nz_in

    nx = nx_in
    nz = nz_in
    dx = dx_in

  endsubroutine

!================================================================
! Functions (for the adjoints, dot product test, etc)
  function shiftL_op(adj,add,model,data) result(stat)
    logical,intent(in)           :: adj,add
    real,dimension(:)          :: model,data
    integer                      :: stat

    call shiftL_op2(adj,add,model,data)
    stat=0
  endfunction

  function shiftU_op(adj,add,model,data) result(stat)
    logical,intent(in)           :: adj,add
    real,dimension(:)          :: model,data
    integer                      :: stat
    call shiftU_op2(adj,add,model,data)
    stat=0
  endfunction
!================================================================

  subroutine model_op(m1_in,m2_in,m3_in,m1_x_in,m1_z_in)

    integer                       :: iz,ix
    real, dimension(:,:),target :: m1_in,m2_in,m3_in,m1_x_in,m1_z_in

    m1 => m1_in
    m2 => m2_in
    m3 => m3_in
    m1_x => m1_x_in
    m1_z => m1_z_in

    !Calculate inverse of density
    !These also calculate the values of density at the velocity staggered grid,
    !using the arithmetic average proposed in Moczo, 2002

    !Rho_vx
    do ix=1,nx-1
      do iz=1,nz
        m1_x(iz,ix)=1.0/(0.5*(m1(iz,ix+1) + m1(iz,ix)))
      enddo
    enddo
    !Rho_vx bottom
    do iz=1,nz
      m1_x(iz,nx)=(1.0/m1(iz,nx))
    enddo
    !Rho_vz
    do ix=1,nx
      do iz=1,nz-1
        m1_z(iz,ix)=1.0/(0.5*(m1(iz+1,ix) + m1(iz,ix)))
      enddo
    enddo
    !Rho_vz right side
    do ix=1,nx
      m1_z(nz,ix)=(1.0/m1(nz,ix))
    enddo


    !write(0,*)"maxval(m1)=",maxval(m1)
    !write(0,*)"maxval(m1_x)=",maxval(m1_x)
    !write(0,*)"maxval(m1_z)=",maxval(m1_z)
    !write(0,*)"minval(m1)=",minval(m1)
    !write(0,*)"minval(m1_x)=",minval(m1_x)
    !write(0,*)"minval(m1_z)=",minval(m1_z)

    !Calculate weigthed average for the lambda variable
    !These put lambda and (lambda+2mi) in their respective staggered grids,
    !using the harmonic average proposed in Moczo, 2002
    do ix=2,nx-1
      do iz=2,nz-1
        !m2(iz,ix)=4.0/((1.0/m2(iz,ix))+(1.0/m2(iz+1,ix))+(1.0/m2(iz,ix+1))+(1.0/m2(iz+1,ix+1)))
        if (m3(iz,ix)==0.or.m3(iz+1,ix)==0.or.m3(iz,ix+1)==0.or.m3(iz+1,ix+1)==0) then
          m3(iz,ix)=0.0
        else
          m3(iz,ix)=4.0/((1.0/m3(iz,ix))+(1.0/m3(iz+1,ix))+(1.0/m3(iz,ix+1))+(1.0/m3(iz+1,ix+1)))
        endif
      enddo
    enddo
    do ix=1,nx
      !m2(1,ix) = m2(2,ix)
      !m2(nz,ix) = m2(nz-1,ix)
      m3(1,ix) = m3(2,ix)
      m3(nz,ix) = m3(nz-1,ix)
    enddo
    do iz=1,nz
      !m2(iz,1) = m2(iz,2)
      !m2(iz,nx) = m2(iz,nx-1)
      m3(iz,1) = m3(iz,2)
      m3(iz,nx) = m3(iz,nx-1)
    enddo

  endsubroutine

!================================================================
  subroutine shiftL_op2(adj,add,model,data)
    integer                     :: iz,ix
    real, dimension(nz,nx),intent(inout) :: model,data
    logical                     :: adj,add

    if (adj) then
      if (.not.add) model = 0.
      do ix=1,nx-1
        do iz=1,nz
          model(iz,ix  ) = model(iz,ix  ) + 0.5*data(iz,ix)
          model(iz,ix+1) = model(iz,ix+1) + 0.5*data(iz,ix)
        enddo
      enddo
    else
      if (.not.add) data = 0.
      do ix=1,nx-1
        do iz=1,nz
          data(iz,ix)= data(iz,ix) + 0.5*(model(iz,ix+1)+model(iz,ix))
        enddo
      enddo
    endif

  endsubroutine

  subroutine shiftU_op2(adj,add,model,data)
    integer                     :: iz,ix
    real, dimension(nz,nx),intent(inout) :: model,data
    logical                     :: adj,add

    if (adj) then
      if (.not.add) model = 0.
      do ix=1,nx
        do iz=1,nz-1
          model(iz  ,ix) = model(iz  ,ix) + 0.5*data(iz,ix)
          model(iz+1,ix) = model(iz+1,ix) + 0.5*data(iz,ix)
        enddo
      enddo
    else
      if (.not.add) data = 0.
      do ix=1,nx
        do iz=1,nz-1
          data(iz,ix)= data(iz,ix) + 0.5*(model(iz+1,ix)+model(iz,ix))
        enddo
      enddo
    endif

  endsubroutine

  subroutine model_delta(nz_in,nx_in,dm2_in,dm3_in)
    integer                     :: iz,ix
    integer                     :: nz_in, nx_in
    real, dimension(:,:),target :: dm2_in,dm3_in

    nx = nx_in
    nz = nz_in
    dm2 => dm2_in
    dm3 => dm3_in

    do ix=2,nx-1
      do iz=2,nz-1
        if (dm2(iz,ix)<1e-2.or.dm2(iz+1,ix)<1e-2.or.dm2(iz,ix+1)<1e-2.or.dm2(iz+1,ix+1)<1e-2) then
          dm2(iz,ix)=0.0
        else
          dm2(iz,ix)=4.0/((1.0/dm2(iz,ix))+(1.0/dm2(iz+1,ix))+(1.0/dm2(iz,ix+1))+(1.0/dm2(iz+1,ix+1)))
        endif
        if (dm3(iz,ix)<1e-2.or.dm3(iz+1,ix)<1e-2.or.dm3(iz,ix+1)<1e-2.or.dm3(iz+1,ix+1)<1e-2) then
          dm3(iz,ix)=0.0
        else
          dm3(iz,ix)=4.0/((1.0/dm3(iz,ix))+(1.0/dm3(iz+1,ix))+(1.0/dm3(iz,ix+1))+(1.0/dm3(iz+1,ix+1)))
        endif
      enddo
    enddo
    do ix=1,nx
      dm2(1,ix) = dm2(2,ix)
      dm2(nz,ix) = dm2(nz-1,ix)
      dm3(1,ix) = dm3(2,ix)
      dm3(nz,ix) = dm3(nz-1,ix)
    enddo
    do iz=1,nz
      dm2(iz,1) = dm2(iz,2)
      dm2(iz,nx) = dm2(iz,nx-1)
      dm3(iz,1) = dm3(iz,2)
      dm3(iz,nx) = dm3(iz,nx-1)
    enddo

  endsubroutine

end module
